@@ -76,7 +76,6 @@ f32 = util.f32
 KeyArray = random.KeyArray
 NeighborListFns = partition.NeighborListFns
 ShiftFn = space.ShiftFn
-UpdateFn = simulate.UpdateFn
 
 
 """Quaternion Utilities.
@@ -443,8 +442,9 @@ def _(position: RigidBody) -> int:
   return tree_reduce(lambda accum, x: accum + x, sizes, 0)
 
 
-@simulate.initialize_momenta.register
-def _(R: RigidBody, mass: RigidBody, key: Array, kT: float):
+@simulate.initialize_momenta.register(RigidBody)
+def _(state, key: Array, kT: float):
+  R, mass = state.position, state.mass
   center_key, angular_key = random.split(key)
 
   P_center = jnp.sqrt(mass.center * kT) * random.normal(center_key,
@@ -468,22 +468,56 @@ def _(R: RigidBody, mass: RigidBody, key: Array, kT: float):
     shape, dtype = R.orientation.shape, R.orientation.dtype
     P_orientation = scale * random.normal(angular_key, shape, dtype=dtype)
 
-  return RigidBody(P_center, P_orientation)
+  return state.set(momentum=RigidBody(P_center, P_orientation))
 
 
-def _rigid_body_3d_update(shift_fn: ShiftFn, m_rot: int) -> UpdateFn:
+class EmptyLeaf:
+  pass
+_EMPTY_LEAF = EmptyLeaf()
+
+
+def split_center_and_orientation(dc):
+  def grab_center(x):
+    if isinstance(x, RigidBody):
+      return x.center
+    return _EMPTY_LEAF
+  def grab_orientation(x):
+    if isinstance(x, RigidBody):
+      return x.orientation
+    return _EMPTY_LEAF
+  def grab_rest(x):
+    if isinstance(x, RigidBody):
+      return _EMPTY_LEAF
+    return x
+  is_rigid = lambda x: isinstance(x, RigidBody)
+  c = tree_map(grab_center, dc, is_leaf=is_rigid)
+  o = tree_map(grab_orientation, dc, is_leaf=is_rigid)
+  r = tree_map(grab_rest, dc, is_leaf=is_rigid)
+  return r, c, o
+
+
+def merge_center_and_orientation(rest, center, orientation):
+  def merge_fn(r, c, o):
+    if r is _EMPTY_LEAF:
+      return RigidBody(c, o)
+    return r
+  return tree_map_no_quat(merge_fn, rest, center, orientation)
+
+
+MOMENTUM_PERMUTATION = [
+  vmap(lambda q: jnp.array([-q[1], q[0], q[3], -q[2]])),
+  vmap(lambda q: jnp.array([-q[2], -q[3], q[0], q[1]])),
+  vmap(lambda q: jnp.array([-q[3], q[2], -q[1], q[0]])),
+]
+
+
+def _rigid_body_3d_position_step(state, shift_fn: ShiftFn, dt, m_rot: int,
+                                 **kwargs):
   """A symplectic update function for 3d rigid bodies."""
-  def com_update_fn(R, P, F, M, dt, **kwargs):
-    return shift_fn(R, P * dt / M, **kwargs)
-
   def free_rotor(k, dt, quat, p_quat, M):
     delta = dt / m_rot
 
-    P = [
-         vmap(lambda q: jnp.array([-q[1], q[0], q[3], -q[2]])),
-         vmap(lambda q: jnp.array([-q[2], -q[3], q[0], q[1]])),
-         vmap(lambda q: jnp.array([-q[3], q[2], -q[1], q[0]])),
-    ]
+    P = MOMENTUM_PERMUTATION
 
     if M.ndim == 1:
       Mk = M[k]
@@ -491,6 +525,7 @@ def _rigid_body_3d_update(shift_fn: ShiftFn, m_rot: int) -> UpdateFn:
       Mk = M[:, [k]]
     else:
       raise NotImplementedError()
+
     zeta = delta * jnp.einsum('ij,ij->i', p_quat, P[k](quat) / (4 * Mk))
     zeta = zeta[:, None]
     quat = jnp.cos(zeta) * quat + jnp.sin(zeta) * P[k](quat)
@@ -498,84 +533,114 @@ def _rigid_body_3d_update(shift_fn: ShiftFn, m_rot: int) -> UpdateFn:
 
     return quat, p_quat
 
-  def quaternion_update_fn(R, P, F, M, dt, **kwargs):
+  def orientation_update(state, dt, **kwargs):
     dt_2 = dt / 2
+    R = state.position
+    P = state.momentum
     if not (isinstance(R, Quaternion) and isinstance(P, Quaternion)):
       raise ValueError('For 3d rigid bodies, orientations must be quaternions.'
                        f'Found {type(R)} for positions and {type(P)} for '
                        'momenta.')
     R = R.vec
     P = P.vec
+    M = state.mass
     for _ in range(m_rot):
       R, P = free_rotor(2, dt_2, R, P, M)
       R, P = free_rotor(1, dt_2, R, P, M)
       R, P = free_rotor(0, dt, R, P, M)
       R, P = free_rotor(1, dt_2, R, P, M)
       R, P = free_rotor(2, dt_2, R, P, M)
-    return Quaternion(R), Quaternion(P)
+    return state.set(position=Quaternion(R), momentum=Quaternion(P))
 
-  def update_fn(R, P, F, M, dt, **kwargs):
-    R_cm = com_update_fn(R.center, P.center, F.center, M.center, dt, **kwargs)
-    R_or, P_or = quaternion_update_fn(
-      R.orientation,
-      P.orientation,
-      F.orientation,
-      M.orientation,
-      dt,
-      **kwargs
-    )
-    R = dataclasses.replace(R, center=R_cm, orientation=R_or)
-    P = dataclasses.replace(P, orientation=P_or)
-    return R, P, F, M
-
-  return update_fn
+  rest, center, orientation = split_center_and_orientation(state)
+  center = simulate.position_step(center, shift_fn, dt, **kwargs)
+  orientation = orientation_update(orientation, dt, **kwargs)
+  return merge_center_and_orientation(rest, center, orientation)
 
 
-def _rigid_body_2d_update(shift_fn: ShiftFn) -> UpdateFn:
+def _rigid_body_2d_position_step(state, shift_fn: ShiftFn, dt,
+                                 **kwargs):
   """A symplectic update function for 2d rigid bodies."""
-  def com_update_fn(R, P, F, M, dt, **kwargs):
-    return shift_fn(R, P * dt / M, **kwargs)
-
-  def orientation_update_fn(R, P, F, M, dt, **kwargs):
-    return R + dt * P / M
-
-  def update_fn(R, P, F, M, dt, **kwargs):
-    R_cm = com_update_fn(R.center, P.center, F.center, M.center, dt, **kwargs)
-    R_or = orientation_update_fn(
-      R.orientation,
-      P.orientation,
-      F.orientation,
-      M.orientation,
-      dt,
-      **kwargs
-    )
-    return dataclasses.replace(R, center=R_cm, orientation=R_or), P, F, M
-
-  return update_fn
+  rest, center, orientation = split_center_and_orientation(state)
+  center = simulate.position_step(center, shift_fn, dt, **kwargs)
+  orientation = simulate.position_step(orientation,
+                                       lambda r, dr, **_: r + dr, dt,
+                                       **kwargs)
+  return merge_center_and_orientation(rest, center, orientation)
 
 
-@simulate.inner_update_fn.register
-def _(R: RigidBody, shift_fn: Callable[..., Array], m_rot=1, **unused_kwargs):
-  if isinstance(R.orientation, Quaternion):
-    return _rigid_body_3d_update(shift_fn, m_rot=m_rot)
+@simulate.position_step.register(RigidBody)
+def _(state, shift_fn, dt, m_rot=1, **kwargs):
+  if isinstance(state.position.orientation, Quaternion):
+    return _rigid_body_3d_position_step(state, shift_fn, dt, m_rot=m_rot,
+                                        **kwargs)
   else:
-    return _rigid_body_2d_update(shift_fn)
+    return _rigid_body_2d_position_step(state, shift_fn, dt, **kwargs)
 
 
-@simulate.canonicalize_mass.register
-def _(mass: RigidBody) -> RigidBody:
+@simulate.stochastic_step.register(RigidBody)
+def _(state, dt: float, kT: float, gamma: float):
+  key, center_key, orientation_key = random.split(state.rng, 3)
+
+  rest, center, orientation = split_center_and_orientation(state)
+
+  center = simulate.stochastic_step(
+    center.set(rng=center_key),
+    dt,
+    kT,
+    gamma.center)
+
+  Pi = orientation.momentum.vec
+  I = orientation.mass
+  G = gamma.orientation
+
+  M = 4 / jnp.sum(1 / I, axis=-1)
+  Q = orientation.position.vec
+  P = MOMENTUM_PERMUTATION
+
+  # First evaluate PI term
+  Pi_mean = 0
+  for l in range(3):
+    I_l = I[:, [l], None]
+    M_l = M[:, None, None]
+    PP = P[l](Q)[:, None, :] * P[l](Q)[:, :, None]
+    Pi_mean += jnp.exp(-G * M_l * dt / (4 * I_l)) * PP
+  Pi_mean = jnp.einsum('nij,nj->ni', Pi_mean, Pi)
+
+  # Then evaluate Q term
+  Pi_var = 0
+  for l in range(3):
+    scale = jnp.sqrt(4 * kT * I[:, l] *
+                     (1 - jnp.exp(-M * G * dt / (2 * I[:, l]))))
+    Pi_var += (scale[:, None] * P[l](Q))**2
+
+  momentum_dist = simulate.Normal(Pi_mean, Pi_var)
+  new_momentum = Quaternion(momentum_dist.sample(orientation_key))
+  orientation = orientation.set(momentum=new_momentum)
+
+  return merge_center_and_orientation(rest.set(rng=key), center, orientation)
+
+
+@simulate.canonicalize_mass.register(RigidBody)
+def _(state):
+  mass = state.mass
   if len(mass.center) == 1:
-    return RigidBody(mass.center[0], mass.orientation)
+    return state.set(mass=RigidBody(mass.center[0], mass.orientation))
   elif len(mass.center) > 1:
-    return RigidBody(mass.center[:, None], mass.orientation)
+    return state.set(mass=RigidBody(mass.center[:, None], mass.orientation))
   raise NotImplementedError(
     'Center of mass must be either a scalar or a vector. Found an array of '
     f'shape {mass.center.shape}.')
 
 
-@simulate.kinetic_energy.register
-def _(R: RigidBody, P: RigidBody, M: RigidBody) -> Array:
-  return kinetic_energy(R, P, M)
+@simulate.kinetic_energy.register(RigidBody)
+def _(state) -> Array:
+  return kinetic_energy(state.position, state.momentum, state.mass)
+
+
+@simulate.temperature.register(RigidBody)
+def _(state) -> Array:
+  return temperature(state.position, state.momentum, state.mass)
 
 
 """Rigid bodies as unions of point-like particles.
@@ -806,7 +871,8 @@ def transform(body: RigidBody, shape: RigidPointUnion) -> Array:
 
 def union_to_points(body: RigidBody,
                     shape: RigidPointUnion,
-                    shape_species: Optional[onp.ndarray]=None
+                    shape_species: Optional[onp.ndarray]=None,
+                    **kwargs,
                     ) -> Tuple[Array, Optional[Array]]:
   """Transforms points in a RigidPointUnion to world space."""
   if shape_species is None:
